// stack.cpp
//
// CompSci 141 / CSE 141 / Informatics 101 Spring 2013
// In-class example: a Stack ADT in C
//
// Contains the implementation details (the functions and the definition of
// the stack struct) for the Stack ADT.  I've implemented the stack as a
// linked list.


// We need stdlib.h for malloc() and free().  We'll also go ahead and include
// stack.h here, so that all of the functions will be declared.  This helps
// prevent us from having to worry about the order in which we define the
// functions; if they're all declared up front, the order of their definition
// in the source file is irrelevant.
#include <stdlib.h>
#include "stack.h"



// Since we're building a linked list based stack, we'll need list nodes.
// The appropriate way in C to define a list node is as a structure.  A
// structure aggregates data, but does not contain methods.
struct stack_node
{
	int value;
	stack_node* next;
};


// Now that we're in the "hidden" part of the Stack ADT, we can state what
// a stack really looks like.  The rest of the world won't have this
// knowledge, which is a good thing.
struct stack
{
	stack_node* top;
	int count;
};


// As a convenience, I've written a little helper function called create_node
// that will allocate a new stack_node and fill in its values, somewhat like
// a constructor might do for us.
stack_node* create_node(int value, stack_node* next)
{
	// The "new" operator allocates an object on the heap and returns a
	// pointer to it.
	stack_node* n = new stack_node;

	// The -> operator that's being used here (and throughout this example)
	// is the C and C++ equivalent of the . operator in Java.  It means
	// "follow this pointer to where it points, then access this member of
	// the thing it points to".  So, this line:
	//
	//     n->value = value;
	//
	// ...assigns the parameter "value" into the value field of the structure
	// that n points to.

	n->value = value;
	n->next = next;
	return n;
}


stack* stack_create()
{
	stack* s = new stack;
	s->top = nullptr;
	s->count = 0;
	return s;
}


void stack_destroy(stack* s)
{
	// Before actually deallocating the stack structure itself, we need
	// to first go through and deallocate all of the stack_nodes, if any.
	// Rather than do this explicitly, we can use the existing stack_size()
	// and stack_pop() functions to do the work for us instead.  (This is
	// a good example of why we want stack.h to be included.  Since all of
	// the functions are declared in stack.h, we don't have to worry about
	// defining stack_size() and stack_pop() before stack_destroy().)
	while (stack_size(s) > 0)
	{
		stack_pop(s);
	}

	// The "delete" operator deallocates the memory block pointed to by the
	// pointer it's given.  (Note that it doesn't delete the pointer, which
	// has been stack-allocated and will die on its own.)
	delete s;
}


void stack_push(stack* s, int i)
{
	s->top = create_node(i, s->top);
	s->count++;
}


void stack_pop(stack* s)
{
	if (s->top != nullptr)
	{
		// It's necessary to delete the stack_node that's being removed
		// from the stack, which is why the temporary pointer and the
		// use of "delete" is necessary here.  (Garbage collectors simplify
		// matters a lot, relative to low-level memory allocation operations
		// like this, though more recent additions to C++ are making this
		// kind of technique gradually less common.)

		stack_node* temp = s->top;
		s->top = s->top->next;
		delete temp;

		s->count--;
	}
}


int stack_top(stack* s)
{
	if (s->top != nullptr)
	{
		return s->top->value;
	}
	else
	{
		// Well, my goodness... Isn't *this* cheesy??  After writing a line
		// of code this bad, I think maybe I need a shower... The reason I
		// did this is that C has no exceptions.  A better design for this ADT,
		// in light of that fact, might be to have a different signature for
		// this method, one that returns an error code rather than the top
		// value, then returns the top value in a pass-by-reference parameter.
		return -999;
	}
}


int stack_size(stack* s)
{
	return s->count;
}
